from unittest import TestCase, main

from .ebpf import EBPF, Instruction
from .bpf import ProgType


class Tests(TestCase):
    def test_assemble(self):
        e = EBPF()
        e.append(0x24, 3, 4, 0x2c3d, 0x2d3e4f5e)
        self.assertEqual(e.assemble(), b"$C=,^O>-")

    def test_assign(self):
        e = EBPF()
        e.r5 = 7
        e.r6 = e.r3
        self.assertEqual(e.opcodes, 
            [Instruction(0xb7, 5, 0, 0, 7),
             Instruction(0xbf, 6, 3, 0, 0)])

    def test_short(self):
        e = EBPF()
        e.s3 = 7
        e.s4 = e.s1
        e.s2 += 3
        e.s5 += e.s6
        self.assertEqual(e.opcodes, 
            [Instruction(0xb4, 3, 0, 0, 7),
             Instruction(0xbc, 4, 1, 0, 0),
             Instruction(opcode=4, dst=2, src=0, off=0, imm=3),
             Instruction(opcode=0xc, dst=5, src=6, off=0, imm=0)])

    def test_augassign(self):
        e = EBPF()
        e.r5 += 7
        e.r3 += e.r6
        e.r4 -= 3
        e.r4 -= e.r7
        e.r4 *= 3
        e.r4 *= e.r7
        e.r4 /= 3
        e.r4 /= e.r7
        e.r4 |= 3
        e.r4 |= e.r7
        e.r4 &= 3
        e.r4 &= e.r7
        e.r4 <<= 3
        e.r4 <<= e.r7
        e.r4 >>= 3
        e.r4 >>= e.r7
        e.r4 %= 3
        e.r4 %= e.r7
        e.r4 ^= 3
        e.r4 ^= e.r7
        e.sr4 >>= 3
        e.sr4 >>= e.r7

        self.assertEqual(e.opcodes, 
            [Instruction(opcode=7, dst=5, src=0, off=0, imm=7),
             Instruction(opcode=15, dst=3, src=6, off=0, imm=0),
             Instruction(opcode=0x17, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x1f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x27, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x2f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x37, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x3f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x47, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x4f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x57, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x5f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x67, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x6f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x77, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x7f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0x97, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0x9f, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0xa7, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0xaf, dst=4, src=7, off=0, imm=0),
             Instruction(opcode=0xc7, dst=4, src=0, off=0, imm=3),
             Instruction(opcode=0xcf, dst=4, src=7, off=0, imm=0)])

    def test_memory(self):
        e = EBPF()
        e.m8[e.r5] = 7
        e.m16[e.r3 + 2] = 3
        e.m32[7 + e.r8] = 5
        e.m64[e.r3 - 7] = 2
        e.m8[e.r5] = e.r1
        e.m16[e.r3 + 2] = e.r2
        e.m32[7 + e.r8] = e.r3
        e.m64[e.r3 - 7] = e.r4
        e.r2 = e.m8[e.r5]
        e.r3 = e.m16[e.r3 + 2]
        e.r4 = e.m32[7 + e.r8]
        e.r5 = e.m64[e.r3 - 7]
        self.assertEqual(e.opcodes,
            [Instruction(opcode=114, dst=5, src=0, off=0, imm=7),
             Instruction(opcode=106, dst=3, src=0, off=2, imm=3),
             Instruction(opcode=98, dst=8, src=0, off=7, imm=5),
             Instruction(opcode=122, dst=3, src=0, off=-7, imm=2),
             Instruction(opcode=115, dst=5, src=1, off=0, imm=0),
             Instruction(opcode=107, dst=3, src=2, off=2, imm=0),
             Instruction(opcode=99, dst=8, src=3, off=7, imm=0),
             Instruction(opcode=123, dst=3, src=4, off=-7, imm=0),
             Instruction(opcode=113, dst=2, src=5, off=0, imm=0),
             Instruction(opcode=105, dst=3, src=3, off=2, imm=0),
             Instruction(opcode=97, dst=4, src=8, off=7, imm=0),
             Instruction(opcode=121, dst=5, src=3, off=-7, imm=0)])


    def test_jump(self):
        e = EBPF()
        target = e.jump()
        e.r0 = 1
        target.target()
        t1 = e.jumpIf(e.r5 > 3)
        t2 = e.jumpIf(e.r1 > e.r2)
        t3 = e.jumpIf(e.r7 >= 2)
        t4 = e.jumpIf(e.r4 >= e.r3)
        e.r0 = 1
        t1.target()
        t2.target()
        t3.target()
        t4.target()
        t1 = e.jumpIf(e.r5 < 3)
        t2 = e.jumpIf(e.r1 < e.r2)
        t3 = e.jumpIf(e.r7 <= 2)
        t4 = e.jumpIf(e.r4 <= e.r3)
        e.r0 = 1
        t1.target()
        t2.target()
        t3.target()
        t4.target()
        t1 = e.jumpIf(e.sr5 > 3)
        t2 = e.jumpIf(e.sr1 > e.sr2)
        t3 = e.jumpIf(e.sr7 >= 2)
        t4 = e.jumpIf(e.sr4 >= e.sr3)
        e.r0 = 1
        t1.target()
        t2.target()
        t3.target()
        t4.target()
        t1 = e.jumpIf(e.sr5 < 3)
        t2 = e.jumpIf(e.sr1 < e.sr2)
        t3 = e.jumpIf(e.sr7 <= 2)
        t4 = e.jumpIf(e.sr4 <= e.sr3)
        e.r0 = 1
        t1.target()
        t2.target()
        t3.target()
        t4.target()
        t1 = e.jumpIf(e.sr5 == 3)
        t2 = e.jumpIf(e.sr1 == e.sr2)
        t3 = e.jumpIf(e.sr7 != 2)
        t4 = e.jumpIf(e.sr4 != e.sr3)
        e.r0 = 1
        t1.target()
        t2.target()
        t3.target()
        t4.target()
        t1 = e.jumpIf(e.sr5 & 3)
        t2 = e.jumpIf(e.sr1 & e.sr2)
        e.r0 = 1
        t1.target()
        t2.target()
        self.assertEqual(e.opcodes,
            [Instruction(opcode=5, dst=0, src=0, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0x25, dst=5, src=0, off=4, imm=3),
             Instruction(opcode=0x2d, dst=1, src=2, off=3, imm=0),
             Instruction(opcode=0x35, dst=7, src=0, off=2, imm=2),
             Instruction(opcode=0x3d, dst=4, src=3, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0xa5, dst=5, src=0, off=4, imm=3),
             Instruction(opcode=0xad, dst=1, src=2, off=3, imm=0),
             Instruction(opcode=0xb5, dst=7, src=0, off=2, imm=2),
             Instruction(opcode=0xbd, dst=4, src=3, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0x65, dst=5, src=0, off=4, imm=3),
             Instruction(opcode=0x6d, dst=1, src=2, off=3, imm=0),
             Instruction(opcode=0x75, dst=7, src=0, off=2, imm=2),
             Instruction(opcode=0x7d, dst=4, src=3, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0xc5, dst=5, src=0, off=4, imm=3),
             Instruction(opcode=0xcd, dst=1, src=2, off=3, imm=0),
             Instruction(opcode=0xd5, dst=7, src=0, off=2, imm=2),
             Instruction(opcode=0xdd, dst=4, src=3, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0x15, dst=5, src=0, off=4, imm=3),
             Instruction(opcode=0x1d, dst=1, src=2, off=3, imm=0),
             Instruction(opcode=0x55, dst=7, src=0, off=2, imm=2),
             Instruction(opcode=0x5d, dst=4, src=3, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1),
             Instruction(opcode=0x45, dst=5, src=0, off=2, imm=3),
             Instruction(opcode=0x4d, dst=1, src=2, off=1, imm=0),
             Instruction(opcode=0xb7, dst=0, src=0, off=0, imm=1)])

    def test_with(self):
        e = EBPF()
        with e.If(e.r2 > 3) as cond:
            e.r2 = 5
        with cond.Else():
            e.r6 = 7
        self.assertEqual(e.opcodes,
            [Instruction(opcode=0xb5, dst=2, src=0, off=2, imm=3),
             Instruction(opcode=0xb7, dst=2, src=0, off=0, imm=5),
             Instruction(opcode=0x5, dst=0, src=0, off=1, imm=0),
             Instruction(opcode=0xb7, dst=6, src=0, off=0, imm=7)])


class KernelTests(TestCase):
    def test_minimal(self):
        e = EBPF(ProgType.XDP, "GPL")
        e.r6 = 7
        target = e.jumpIf(e.r1 > 3)
        e.r1 = 3
        target.target()
        e.r0 = 0
        e.exit()
        self.assertEqual(e.load(log_level=1), "")


if __name__ == "__main__":
    main()
